import re
import os
import os.path
import sys
import math
from shutil import copyfile

from django.utils.translation import ngettext, gettext as _, ngettext

from galeria.settings import GALERIA_PREFIX, EXPORT_WIDTH, LANGUAGE_CODE, \
    THUMB_MAX, VIEW_MAX, DEFAULT_LATITUDE, DEFAULT_LONGITUDE, STATIC_PATH_PREFIX

from datetime import datetime, timezone, timedelta
import pyexiv2
from PIL import Image
import unicodedata

from babel.dates import format_datetime

ORIENTATION_REVERSE_LANDSCAPE = 3
ORIENTATION_PORTRAIT          = 6
ORIENTATION_REVERSE_PORTRAIT  = 8

DEFAULT_GPS_TAG = 9208
DEFAULT_GPS_VERSION = "2 0 0 0"
DEFAULT_GPS_DATUM = "WGS-84"

DATE_TIME = 'Exif.Photo.DateTimeOriginal'
TIME_ZONE = 'Exif.Image.TimeZoneOffset'
ORIENTATION = 'Exif.Image.Orientation'
THUMB_ORIENTATION = 'Exif.Thumbnail.Orientation'
SUBJECT = 'Xmp.dc.subject'
CAPTION = 'Iptc.Application2.Caption'
DESCRIPTION = 'Exif.Image.ImageDescription'
LATITUDE = 'Exif.GPSInfo.GPSLatitude'
LONGITUDE = 'Exif.GPSInfo.GPSLongitude'
LATITUDE_REF = 'Exif.GPSInfo.GPSLatitudeRef'
LONGITUDE_REF = 'Exif.GPSInfo.GPSLongitudeRef'
GPS_TAG = 'Exif.Image.GPSTag'
GPS_VERSION = 'Exif.GPSInfo.GPSVersionID'
GPS_DATUM = 'Exif.GPSInfo.GPSMapDatum'
GPS_DATE = 'Exif.GPSInfo.GPSDateStamp'
GPS_TIME = 'Exif.GPSInfo.GPSTimeStamp'

HTACCESS = '.htaccess'
HTA_FILES = '<Files "'
HTA_CLOSE = '">\n'
HTA_ENTRY = ('<Files "FILENAME">\n' \
             '  Require expr "%{QUERY_STRING} =~ m#token=TOKEN#"\n' \
             '</Files>\n')

def normalize(string):
    n = unicodedata.normalize('NFKD', string) \
                   .encode('ASCII', 'ignore') \
                   .decode()                  \
                   .lower()                   \
                   .replace(' ', '-')
    n = re.sub(r'[^\w-]+', '', n)
    return n

def _load_restricted(htaccess):
    restricted = {}
    if not os.path.exists(htaccess):
        return restricted
    with open(htaccess, 'r') as hta:
        for line in hta.readlines():
            if not line.startswith(HTA_FILES):
                continue
            fn = line.replace(HTA_FILES, '') \
                     .replace(HTA_CLOSE, '')
            restricted[fn] = fn
    return restricted

def restrict(items, token):
    directories = {}
    for item in items:
        if hasattr(item, 'location'):
            dn = os.path.dirname(item.location)
            if dn not in directories:
                directories[dn] = []
            directories[dn].append(item.location)
    to_add = {}
    for dn in directories.keys():
        to_add[dn] = []
        htaccess = os.path.join(dn, HTACCESS)
        restricted = _load_restricted(htaccess)
        for fn in directories[dn]:
            if os.path.basename(fn) in restricted:
                continue
            to_add[dn].append(os.path.basename(fn))
    if len(to_add) == 0:
        return
    for sub in [ 'photos', 'exports', 'views', 'thumbs' ]:
        for dn, fns in to_add.items():
            htaccess = os.path.join(dn, HTACCESS) \
                              .replace('/photos/', '/%s/' % sub, 1)
            mode = 'w'
            if os.path.exists(htaccess):
                mode = 'a'
            with open(htaccess, mode) as hta:
                for fn in sorted(fns):
                    hta.write(HTA_ENTRY.replace('FILENAME', fn) \
                              .replace('TOKEN', token.key))
            os.chmod(htaccess, 0o660)

def _free_photo(photo, token):
    dn = os.path.dirname(photo.location)
    htaccess = os.path.join(dn, HTACCESS)
    restricted = _load_restricted(htaccess)
    if len(restricted) == 0:
        return
    fn = os.path.basename(photo.location)
    if fn not in restricted:
        return
    if len(restricted) == 1:
        _delete_htaccess(dn)
        return
    with open(htaccess, 'w') as hta:
        for r in sorted(restricted.values()):
            if r == fn:
                continue
            hta.write(HTA_ENTRY.replace('FILENAME', r) \
                      .replace('TOKEN', token.key))
    os.chmod(htaccess, 0o660)
    for sub in [ 'exports', 'views', 'thumbs' ]:
        copyfile(htaccess, htaccess.replace('/photos/', '/%s/' % sub))

def _delete_htaccess(path):
    for sub in [ 'photos', 'exports', 'views', 'thumbs' ]:
        htaccess = os.path.join(path, HTACCESS) \
                          .replace('/photos/', '/%s/' % sub, 1)
        os.remove(htaccess)

def _free_album(album):
    if album.the_parent and album.the_parent.the_parent:
        path = os.path.join(STATIC_PATH_PREFIX, 'galeria', 'photos',
                            album.the_parent.the_parent.slug,
                            album.the_parent.slug, album.slug)
        _delete_htaccess(path)
    for subalbum in album.get_subalbums():
        _free_album(subalbum)

def free(item, token):
    if not item:
        return
    if hasattr(item, 'location'):
        _free_photo(item, token)
    else:
        _free_album(item)

def fractions2float(fractions):
    i = 0
    d = [ 0.0, 0.0, 0.0]
    for fraction in fractions:
        d[i] = float(fraction.numerator) / float(fraction.denominator)
        i += 1
    return d[0] + d[1] / 60.0 + d[2] / 3600.0

def get_month_name(ts, offset=0, capitalize=False):
    tz = timezone(timedelta(hours=offset))
    locale = LANGUAGE_CODE.replace('-', '_')
    name = format_datetime(ts, 'MMMM', locale=locale, tzinfo=tz)
    if capitalize:
        name = name.capitalize()
    return name

def get_year_name(year):
    return _('Year %(year)s') % { 'year': year }

def get_image_metadata(path):
    image = Image.open(path)
    if image is None:
        return None
    w, h = image.size
    metadata = pyexiv2.metadata.ImageMetadata(path)
    metadata.read()

    transpose = None
    try:
        if ORIENTATION in metadata:
            o = metadata.get(ORIENTATION).value
            if o == ORIENTATION_REVERSE_LANDSCAPE:
                transpose = Image.ROTATE_180
            if o == ORIENTATION_PORTRAIT:
                transpose = Image.ROTATE_270
            if o == ORIENTATION_REVERSE_PORTRAIT:
                transpose = Image.ROTATE_90
    except Exception as e:
        sys.stderr.write('Error getting tags: %s\n' % (str(e)))
        return None
    if transpose == Image.ROTATE_270 or transpose == Image.ROTATE_90:
        w, h = h, w

    md = { 'image': image, 'width': w, 'height': h }
    md['export_width'] = EXPORT_WIDTH
    md['export_height'] = int((h * EXPORT_WIDTH)/w)
    if w > h:
        md['view_width'] = VIEW_MAX
        md['view_height'] = int((h * VIEW_MAX)/w)
        md['thumb_width'] = THUMB_MAX
        md['thumb_height'] = int((h * THUMB_MAX)/w)
    else:
        md['view_height'] = VIEW_MAX
        md['view_width'] = int((w * VIEW_MAX)/h)
        md['thumb_height'] = THUMB_MAX
        md['thumb_width'] = int((w * THUMB_MAX)/h)

    try:
        if CAPTION in metadata:
            md['title'] = metadata.get(CAPTION).value[0]
        else:
            md['title'] = os.path.basename(path)
            if '.' in md['title']:
                t = os.path.basename(path).split('.')[:-1]
                md['title'] = os.path.join(*t)
        if DESCRIPTION in metadata:
            md['description'] = metadata.get(DESCRIPTION).value
        else:
            md['description'] = ''
        if TIME_ZONE in metadata:
            md['time_offset'] = int(metadata.get(TIME_ZONE).value)
        else:
            md['time_offset'] = 0
        tz = timezone(timedelta(hours=md['time_offset']))
        if DATE_TIME in metadata:
            dt = metadata.get(DATE_TIME).value
        else:
            dt = datetime.fromtimestamp(os.path.getmtime(path))
        md['timestamp'] = datetime(dt.year, dt.month, dt.day, dt.hour,
                                   dt.minute, dt.second, dt.microsecond, tz)
        if SUBJECT in metadata:
            v = metadata.get(SUBJECT).value
            if len(v) > 0:
                md['album'] = v[0]
        if 'album' not in md:
            locale = LANGUAGE_CODE.replace('-', '_')
            md['album'] = format_datetime(md['timestamp'], 'eeee d',
                                          locale=locale).capitalize()
        if LATITUDE in metadata and LATITUDE_REF in metadata and \
           LONGITUDE in metadata and LONGITUDE_REF in metadata:
            lat = metadata.get(LATITUDE).value
            lon = metadata.get(LONGITUDE).value
            latitude = fractions2float(lat)
            longitude = fractions2float(lon)
            if metadata.get(LATITUDE_REF).value == 'S':
                latitude *= -1.0
            if metadata.get(LONGITUDE_REF).value == 'W':
                longitude *= -1.0
            md['latitude'] = latitude
            md['longitude'] = longitude
        else:
            md['latitude'] = DEFAULT_LATITUDE
            md['longitude'] = DEFAULT_LONGITUDE
        md['location'] = os.path.join(GALERIA_PREFIX,
                                      '%04d' % md['timestamp'].year,
                                      '%02d' % md['timestamp'].month,
                                      normalize(md['album']),
                                      os.path.basename(path))
        if os.path.exists(md['location']):
            sys.stderr.write('Photo already exists: %s\n' % md['location'])
            return None

        umask = os.umask(0o007)
        photo_dir = os.path.dirname(md['location'])
        if not os.path.exists(photo_dir):
            os.makedirs(photo_dir, mode=0o770)
        export_dir = photo_dir.replace('photos', 'exports')
        if not os.path.exists(export_dir):
            os.makedirs(export_dir, mode=0o770)
        view_dir = photo_dir.replace('photos', 'views')
        if not os.path.exists(view_dir):
            os.makedirs(view_dir, mode=0o770)
        thumb_dir = photo_dir.replace('photos', 'thumbs')
        if not os.path.exists(thumb_dir):
            os.makedirs(thumb_dir, mode=0o770)
        os.umask(umask)

        copyfile(path, md['location'])
        os.chmod(md['location'], 0o660)

        if transpose:
            image = image.transpose(transpose)
        dimension = md['export_width'], md['export_height']
        image.thumbnail(dimension, Image.ANTIALIAS)
        export = md['location'].replace('photos', 'exports')
        image.save(export, 'JPEG')
        os.chmod(export, 0o660)

        image = Image.open(path)
        if transpose:
            image = image.transpose(transpose)
        dimension = md['view_width'], md['view_height']
        image.thumbnail(dimension, Image.ANTIALIAS)
        view = md['location'].replace('photos', 'views')
        image.save(view, 'JPEG')
        os.chmod(view, 0o660)

        image = Image.open(path)
        if transpose:
            image = image.transpose(transpose)
        dimension = md['thumb_width'], md['thumb_height']
        image.thumbnail(dimension, Image.ANTIALIAS)
        thumb = md['location'].replace('photos', 'thumbs')
        image.save(thumb, 'JPEG')
        os.chmod(thumb, 0o660)
    except Exception as e:
        sys.stderr.write('Error getting tags: %s\n' % (str(e)))
        return None

    return md
